# coding:utf-8
import json

from ks3.exception import S3ResponseError
from ks3.http import make_request_v4


def assumeRole(access_key_id, access_key_secret, role_krn, role_session_name, duration_seconds=3600, policy=None,
               inner_api=False):
    query_args = {
        'Action': 'AssumeRole',
        'Version': '2015-11-01'
    }
    headers = {
        'content-type': 'application/json',
        'accept': 'application/json'
    }
    body = {
        "RoleKrn": role_krn,  # "krn:ksc:iam::xxx:role/xxx", 角色id
        "RoleSessionName": role_session_name,  # "KS3", 该次扮演的名称
        "DurationSeconds": duration_seconds  # 持续有效时间，非必须参数，默认3600秒
    }
    if policy is not None:
        body['Policy'] = json.dumps(policy)

    response = make_request_v4(access_key_id, access_key_secret, method='GET', service='sts', region='cn-beijing-6',
                               headers=headers,
                               query_args=query_args, body=json.dumps(body), inner_api=inner_api)
    body = response.read()
    if response.status != 200:
        raise S3ResponseError(response.status, response.reason, body)
    return json.loads(body)



__version__ = '1.12.1'

import logging
from logging.handlers import RotatingFileHandler


def set_file_logger(
        file_name, name='ks3', level=logging.INFO,
        format_string=None, max_bytes=1024 * 1024 * 500, backup_count=5
):
    if not format_string:
        format_string = '%(asctime)s %(name)s:%(lineno)s [%(levelname)s] %(thread)d - %(funcName)s: %(message)s'
    logger = logging.getLogger(name)
    logger.setLevel(level)
    fh = RotatingFileHandler(filename=file_name, maxBytes=max_bytes, backupCount=backup_count)
    fh.setLevel(level)
    formatter = logging.Formatter(format_string)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


def set_stream_logger(name='ks3', level=logging.INFO, format_string=None):
    if not format_string:
        format_string = '%(asctime)s %(name)s:%(lineno)s [%(levelname)s] %(thread)d - %(funcName)s: %(message)s'
    logger = logging.getLogger(name)
    logger.setLevel(level)
    sh = logging.StreamHandler()
    sh.setLevel(level)
    formatter = logging.Formatter(format_string)
    sh.setFormatter(formatter)
    logger.addHandler(sh)



# coding:utf-8
import logging
import time
import re
import os

from requests.adapters import DEFAULT_POOLBLOCK
from requests import Request
from urllib3 import poolmanager

from ks3 import utils
import select
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPConnection, VerifiedHTTPSConnection
from urllib3.connectionpool import HTTPConnectionPool, HTTPSConnectionPool
from ks3.utils import get_default_user_agent

try:
    import http.client as httpcli  # for Python 3
    import urllib.parse as parse
    from urllib import parse as urlparse
except ImportError:
    print('python@2.x no longer supported by ks3sdk@2.x, please refer to ks3sdk@1.x for python@2.x support')

from ks3.auth import canonical_string, add_auth_header, url_encode, encode
from ks3.authV4 import add_auth_header as add_auth_header_v4

logger = logging.getLogger(__name__)


class CallingFormat:
    PATH = 1
    SUBDOMAIN = 2
    VANITY = 3


class AuthingFormat:
    V2 = 1
    V4 = 2


def merge_meta(headers, metadata):
    final_headers = headers.copy()
    for k in list(metadata.keys()):
        final_headers["x-kss-" + "meta-" + k] = metadata[k]

    return final_headers


def query_args_hash_to_string(query_args):
    pairs = []
    for k, v in list(query_args.items()):
        piece = k
        if v is not None:
            piece += "=%s" % parse.quote_plus(str(v).encode('utf-8'))
            # piece += "=%s" % v
        pairs.append(piece)

    return '&'.join(pairs)


def get_object_url(age, bucket="", key="", secret_access_key="", access_key_id="", query_args={}):
    expire = str(int(time.time()) + age)
    headers = {"Date": expire}
    c_string = canonical_string("GET", bucket, key, query_args, headers)
    path = c_string.split("\n")[-1]

    signature = parse.quote_plus(encode(secret_access_key, c_string))
    if "?" in path:
        url = "http://kss.ksyun.com%s&Expires=%s&AccessKeyId=%s&Signature=%s" % \
              (path, expire, access_key_id, signature)
    else:
        url = "http://kss.ksyun.com%s?Expires=%s&AccessKeyId=%s&Signature=%s" % \
              (path, expire, access_key_id, signature)
    return url


def get_proxy_url(scheme, host, port, proxy_username, proxy_password):
    if host is None:
        return None
    if port is None:
        address = host
    else:
        address = '{0}:{1}'.format(host, port)
    if proxy_username is not None and proxy_password is not None:
        proxy_auth = '{0}:{1}@'.format(proxy_username, proxy_password)
    else:
        proxy_auth = ''
    return {scheme: 'http://{0}{1}'.format(proxy_auth, address)}


def do_request(session, method, url, headers=None, data=None, proxy=None, timeout=10):
    headers = headers or {}
    data = data or {}
    proxy = proxy or {}

    req = Request(
        method=method.upper(),
        url=url,
        headers=headers,
        data=data,
    )
    prep = session.prepare_request(req)
    prep.headers.update(headers)
    stream = True
    verify = True
    settings = session.merge_environment_settings(
        prep.url, proxy, stream, verify, None
    )
    send_kwargs = {
        "timeout": timeout,
        "allow_redirects": False,
    }
    send_kwargs.update(settings)
    resp = session.send(prep, **send_kwargs)
    resp = KS3Response(resp)
    return resp


def make_request(server, port, access_key_id, access_key_secret, bucket="", key="", query_args=None, headers=None,
                 data="", metadata=None, method="PUT", calling_format=None, is_secure=False,
                 domain_mode=False, need_auth_header=True, timeout=10, ua_addon='', proxy_host=None,
                 proxy_port=None, proxy_username=None, proxy_password=None, session=None):
    if not headers:
        headers = {}
    # if not query_args:
    #    query_args = {}
    if not metadata:
        metadata = {}

    if bucket and not domain_mode:
        server = calling_format.get_bucket_server(server, bucket)
    path = calling_format.build_path_base(bucket, key)
    # path += "/%s" % url_encode(key)
    # path = path.replace('//', '/%2F')

    if query_args:
        if isinstance(query_args, dict):
            path += "?" + query_args_hash_to_string(query_args)
        else:
            path += "?" + query_args

    host = "%s:%d" % (server, port)

    headers['User-Agent'] = get_default_user_agent() + ' ' + ua_addon
    final_headers = merge_meta(headers, metadata)
    if method == "PUT" and "Content-Length" not in final_headers and not data:
        final_headers["Content-Length"] = "0"
    if method.upper() == "POST" and "Content-Length" not in final_headers and not data:
        final_headers["Content-Length"] = str(len(data))
    if need_auth_header:
        add_auth_header(access_key_id, access_key_secret, final_headers, method,
                        bucket, key, query_args)
    final_headers['Accept-Encoding'] = 'identity'
    logger.info('send [{method}] request, host: {host}, port: {port}, path: {path}, headers: {headers}'
                .format(method=method, host=host, port=port, path=path, headers=final_headers))

    scheme = 'https' if is_secure else 'http'
    proxy = get_proxy_url(scheme, proxy_host, proxy_port, proxy_username, proxy_password)

    resp = do_request(
        session=session,
        method=method,
        url='{scheme}://{host}{path}'.format(scheme=scheme, host=host, path=path),
        headers=final_headers,
        data=data,
        proxy=proxy,
        timeout=timeout
    )

    logger.info(
        'complete [{method}] request, host: {host}, port: {port}, path: {path}, request_id: {request_id}, status_code:{status}'
        .format(
            method=method,
            host=host,
            port=port,
            path=path,
            request_id=resp.getheader('x-kss-request-id') if resp else '',
            status=resp.status,
        )
    )
    if 300 <= resp.status < 400:
        loc = resp.getheader('location')
        if loc:
            reg = re.findall(r'http[s]?://(.*?)(:\d+)?/', loc)
            if reg:
                # 如果返回的是bucket style域名，需要提取region域名出来
                if len(reg[0][0].split('.')) == 4:
                    new_server = reg[0][0].split('.', 1)[1]
                else:
                    new_server = reg[0][0]
                loc_parse = urlparse.urlparse(loc)
                if 'Signature' in loc_parse.query:
                    try:
                        resp_temp = do_request(
                            session=session,
                            method=method,
                            url='{scheme}://{host}{path}'.format(
                                scheme=scheme, host=new_server,
                                path=loc_parse.path + '?' + loc_parse.query
                            ),
                            headers=final_headers,
                            data=data,
                            proxy=proxy,
                            timeout=timeout
                        )
                        return resp_temp
                    except Exception as err:
                        print(str(err))
                else:
                    if hasattr(data, 'read'):
                        data.seek(0, os.SEEK_SET)
                        if isinstance(data, utils.FpAdapter):
                            data.reset_crc_process()
                    return make_request(new_server, port, access_key_id, access_key_secret, bucket, key, query_args,
                                        headers, data, metadata, method=method, calling_format=calling_format,
                                        is_secure=is_secure, domain_mode=domain_mode,
                                        need_auth_header=True, timeout=timeout, ua_addon=ua_addon,
                                        proxy_host=proxy_host, proxy_port=proxy_port, session=session)
    return resp


# 发送awsv4的请求
def make_request_v4(access_key_id, access_key_secret, method='', service='', region='', query_args=None, headers={},
                    body="", is_secure=False, timeout=10, inner_api=False):
    inner_string = '.inner' if inner_api else ''
    host = service + inner_string + '.api.ksyun.com'

    if (is_secure):
        connection = httpcli.HTTPSConnection(host)
    else:
        connection = httpcli.HTTPConnection(host)
    connection.timeout = timeout

    path = "/"
    if query_args:
        if isinstance(query_args, dict):
            query_args = query_args_hash_to_string(query_args)
    path += "?" + query_args

    headers = add_auth_header_v4(access_key_id, access_key_secret, region, service, host, method, query_args, body,
                                 headers)

    connection.request(method, path, body, headers)
    resp = connection.getresponse()
    return resp


class KS3Response:
    def __init__(self, resp):
        self.raw_resp = resp  # requests.models.Response
        self.status = resp.status_code
        self.headers = self.msg = resp.raw.headers
        self.reason = resp.reason

    def __bool__(self):
        return self.raw_resp is not None

    def getheader(self, name, default=None):
        return self.headers.get(name, default)

    def getheaders(self):
        return self.headers

    def read(self, amt=None):
        # urllib3.response.HTTPResponse.read()
        return self.raw_resp.raw.read(amt)


def repair_pool_key():
    if 'key_blocksize' in poolmanager.PoolKey._fields:
        return
    try:
        key_fields_tuple = poolmanager._key_fields
        key_list = list(key_fields_tuple)
        key_list.append('key_blocksize')
        patched_key_fields = tuple(key_list)

        import collections
        import functools

        PatchedPoolKey = collections.namedtuple("PatchedPoolKey", patched_key_fields)
        poolmanager.PoolKey = PatchedPoolKey
        poolmanager.key_fn_by_scheme = {
            "http": functools.partial(poolmanager._default_key_normalizer, PatchedPoolKey),
            "https": functools.partial(poolmanager._default_key_normalizer, PatchedPoolKey),
        }
    except AttributeError:
        pass


class KS3HTTPAdapter(HTTPAdapter):
    def __init__(self, block_size=8 * 1024, *args, **kwargs):
        self.block_size = block_size

        repair_pool_key()
        poolmanager.pool_classes_by_scheme = {
            'http': KS3HTTPConnectionPool,
            'https': KS3HTTPSConnectionPool,
        }
        super().__init__(*args, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=DEFAULT_POOLBLOCK, **pool_kwargs):
        if 'key_blocksize' in poolmanager.PoolKey._fields:
            pool_kwargs['blocksize'] = self.block_size
        super().init_poolmanager(connections, maxsize, block=block, **pool_kwargs)


class KS3HTTPConnection(HTTPConnection):
    def _send_output(self, message_body=None, encode_chunked=False):
        """Send the currently buffered request and clear the buffer.

        Appends an extra \\r\\n to the buffer.
        A message_body may be specified, to be appended to the request.
        """
        self._buffer.extend((b"", b""))
        msg = b"\r\n".join(self._buffer)
        del self._buffer[:]
        self.send(msg)
        if message_body is not None:

            # create a consistent interface to message_body
            if hasattr(message_body, 'read'):
                # Let file-like take precedence over byte-like.  This
                # is needed to allow the current position of mmap'ed
                # files to be taken into account.
                chunks = self._read_readable(message_body)
            else:
                try:
                    # this is solely to check to see if message_body
                    # implements the buffer API.  it /would/ be easier
                    # to capture if PyObject_CheckBuffer was exposed
                    # to Python.
                    memoryview(message_body)
                except TypeError:
                    try:
                        chunks = iter(message_body)
                    except TypeError:
                        raise TypeError("message_body should be a bytes-like "
                                        "object or an iterable, got %r"
                                        % type(message_body))
                else:
                    # the object implements the buffer interface and
                    # can be passed directly into socket methods
                    chunks = (message_body,)
            for chunk in chunks:
                ready_to_read, ready_to_write, _ = select.select([self.sock], [self.sock], [], self.timeout)
                if ready_to_read:
                    return
                if not chunk:
                    if self.debuglevel > 0:
                        print('Zero length chunk ignored')
                    continue

                if ready_to_write:
                    try:
                        self.send(chunk)
                    except Exception as e:
                        pos = None
                        if message_body != b'' and hasattr(message_body, 'tell'):
                            pos = message_body.tell()
                        # 有可能写的时候，发生了BrokenPipeError；为了读取到响应的内容，最后再尝试读取一次
                        if isinstance(e, (BrokenPipeError, ConnectionAbortedError, ConnectionResetError)):
                            last_read_ready, _, _ = select.select([self.sock], [self.sock], [], self.timeout)
                            if last_read_ready:
                                return
                            else:
                                logger.error('send chunk error, message body read position={pos}, error: {e}'
                                             .format(pos=pos, e=str(e)),
                                             exc_info=True)
                                raise e
                        else:
                            logger.error('send chunk error, message body read position={pos}, error: {e}'
                                         .format(pos=pos, e=str(e)),
                                         exc_info=True)
                            raise e


class KS3HTTPSConnection(VerifiedHTTPSConnection, KS3HTTPConnection):
    ...


class KS3HTTPConnectionPool(HTTPConnectionPool):
    ConnectionCls = KS3HTTPConnection


class KS3HTTPSConnectionPool(HTTPSConnectionPool):
    ConnectionCls = KS3HTTPSConnection



from ks3.user import User

class DeleteMarker(object):
    def __init__(self, bucket=None, name=None):
        self.bucket = bucket
        self.name = name
        self.version_id = None
        self.is_latest = False
        self.last_modified = None
        self.owner = None

    def startElement(self, name, attrs, connection):
        if name == 'Owner':
            self.owner = User(self)
            return self.owner
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Key':
            self.name = value
        elif name == 'IsLatest':
            if value == 'true':
                self.is_latest = True
            else:
                self.is_latest = False
        elif name == 'LastModified':
            self.last_modified = value
        elif name == 'Owner':
            pass
        elif name == 'VersionId':
            self.version_id = value
        else:
            setattr(self, name, value)



# -*- coding: utf-8 -*-
from ks3.responseResult import ResponseMetadata


class BucketMirror(object):
    def __init__(self, use_default_robots=None, async_mirror_rule=None, sync_mirror_rules=None, *args, **kwargs):
        self.version = "V3"
        self.use_default_robots = use_default_robots
        self.async_mirror_rule = async_mirror_rule
        self.sync_mirror_rules = sync_mirror_rules

        self.response_metadata = ResponseMetadata(**kwargs)

    def keys(self):
        return ('version', 'use_default_robots', 'async_mirror_rule', 'sync_mirror_rules')

    def __getitem__(self, item):
        if item == 'async_mirror_rule':
            return dict(getattr(self, item))
        if item == 'sync_mirror_rules':
            rules_dict = []
            for value in getattr(self, item):
                rules_dict.append(dict(value))
            return rules_dict
        return getattr(self, item)


class SavingSetting(object):
    def __init__(self, acl=None):
        self.acl = acl


class AsyncMirrorRule(object):
    def __init__(self, mirror_urls=None, saving_setting=None):
        if mirror_urls is None:
            mirror_urls = []
        self.mirror_urls = mirror_urls
        self.saving_setting = saving_setting

    @classmethod
    def rule_with_acl(cls, mirror_urls=None, saving_setting_acl=None):
        if mirror_urls is None:
            mirror_urls = []
        saving_setting = SavingSetting(acl=saving_setting_acl)
        return cls(mirror_urls=mirror_urls, saving_setting=saving_setting)

    def keys(self):
        return ('mirror_urls', 'saving_setting')

    def __getitem__(self, item):
        if item == 'saving_setting':
            return getattr(self, item).__dict__
        return getattr(self, item)


class SyncMirrorRules(object):
    def __init__(self, match_condition=None, mirror_url=None, mirror_request_setting=None, saving_setting=None):
        self.saving_setting = saving_setting
        self.mirror_request_setting = mirror_request_setting
        self.mirror_url = mirror_url
        self.match_condition = match_condition

    @classmethod
    def rules_with_prefix_acl(cls, key_prefixes=None, mirror_url=None, mirror_request_setting=None, saving_setting_acl=None):
        saving_setting = SavingSetting(acl=saving_setting_acl)
        match_condition = MatchCondition(key_prefixes=key_prefixes)
        return cls(match_condition=match_condition, mirror_url=mirror_url, mirror_request_setting=mirror_request_setting, saving_setting=saving_setting)

    def keys(self):
        return ('saving_setting', 'mirror_request_setting', 'mirror_url', 'match_condition')

    def __getitem__(self, item):
        if item == 'mirror_request_setting':
            return dict(getattr(self, item))
        if item == 'saving_setting' or item == 'match_condition':
            return getattr(self, item).__dict__
        return getattr(self, item)

class MirrorRequestSetting(object):
    def __init__(self, pass_query_string=None, follow3xx=None, header_setting=None):
        self.pass_query_string = pass_query_string
        self.follow3xx = follow3xx
        self.header_setting = header_setting

    def keys(self):
        return ('pass_query_string', 'follow3xx', 'header_setting')

    def __getitem__(self, item):
        if item == 'header_setting':
            return dict(getattr(self, item))
        return getattr(self, item)

class HeaderSetting(object):
    def __init__(self, set_headers=None, remove_headers=None, pass_all=None, pass_headers=None):
        if pass_headers is None:
            pass_headers = []
        if remove_headers is None:
            remove_headers = []
        if set_headers is None:
            set_headers = []
        self.set_headers = set_headers
        self.remove_headers = remove_headers
        self.pass_all = pass_all
        self.pass_headers = pass_headers

    def keys(self):
        return ('set_headers', 'remove_headers', 'pass_all', 'pass_headers')

    def __getitem__(self, item):
        return getattr(self, item)

# 回源条件
class MatchCondition(object):
    def __init__(self, http_codes=None, key_prefixes=""):
        if http_codes is None:
            http_codes = ["404"]
        self.http_codes = http_codes
        self.key_prefixes = key_prefixes


# -*- coding: utf-8 -*-
# python3
from ks3.connection import Connection
import os

ak = os.getenv('KS3_TEST_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
sk = os.getenv('KS3_TEST_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
bucket_name = os.getenv('KS3_TEST_BUCKET', '<KS3_TEST_BUCKET>')
endpoint = os.getenv('KS3_TEST_ENDPOINT', 'ks3-cn-shanghai-internal.ksyuncs.com')

conn = Connection(ak, sk, host=endpoint, is_secure=False, ua_addon='test-ua/1') #port=8091,
key_name = 'test_key'

customer_key = bytes(32)
import base64, hashlib
customer_key_base64 = base64.b64encode(customer_key).decode('utf-8')
md5 = hashlib.md5()
md5.update(customer_key)
customer_key_md5 = base64.b64encode(md5.digest()).decode('utf-8')

customer_encryption_headers={
  'x-kss-server-side-encryption-customer-algorithm': 'AES256',
  'x-kss-server-side-encryption-customer-key': customer_key_base64,
  'x-kss-server-side-encryption-customer-key-MD5': customer_key_md5
}

customer_key2 = bytearray(b'00000000000000000000000000000000')
customer_key2[0] = 1
print(customer_key2)
import base64, hashlib
customer_key2_base64 = base64.b64encode(customer_key2).decode('utf-8')
md5 = hashlib.md5()
md5.update(customer_key2)
customer_key2_md5 = base64.b64encode(md5.digest()).decode('utf-8')

#####################  ks3.object  ###########################
def getObjectMeta(bucket_name, object_key_name, headers=None):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.get_key_meta(object_key_name, headers=headers)
  if resp:
    print("获取文件header成功: ", resp.headers)

def uploadObjectFromFile(path, file_name=None, encrypt_key=False, headers=None):
  bucket = conn.get_bucket(bucket_name)
  filename = file_name if file_name else path.split('/')[-1]
  k = bucket.new_key(filename)
  ret = k.set_contents_from_filename(path, encrypt_key=encrypt_key, headers=headers)
  if ret and ret.status == 200:
    print("上传成功: %s" % path)

def upload_object_with_customer_encryption_key(key_name, content):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key(key_name)
  ret = k.set_contents_from_string(content, headers=customer_encryption_headers)
  if ret and ret.status == 200:
    print("上传成功: %s" % key_name)

  getObjectMeta(bucket_name, key_name, headers=customer_encryption_headers)
  downloadObjectAndPrint(bucket_name, key_name, headers=customer_encryption_headers)
  bucket.delete_key(key_name)
  print("清理文件成功: %s" % key_name)

def downloadObjectAndPrint(bucket_name, key_name, headers=None):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key(key_name, headers=headers)
  s = k.get_contents_as_string(headers=headers).decode()
  print(s)

def downloadObjectAndSave(key_name, file_save_path, headers=None):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key(key_name, headers=headers)
  k.get_contents_to_filename(file_save_path, headers=headers)

def copy_encryption(dstKey, srcKey, encrypt_key=None):
  b = conn.get_bucket(bucket_name)
  b.copy_key(dstKey, bucket_name, srcKey, encrypt_key=encrypt_key)

def copy_customer_encryption(dstKey, srcKey, encrypt_key=None, headers=None):
  b = conn.get_bucket(bucket_name)
  src_bucket_name = bucket_name
  b.copy_key(dstKey, src_bucket_name, srcKey, encrypt_key=encrypt_key, headers=headers)

def test_copy_customer_encryption():
  srcKey = 'test-customer-encryption'
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key(srcKey)
  ret = k.set_contents_from_string('hello customer encryption', headers=customer_encryption_headers)
  if ret and ret.status == 200:
    print("上传成功: %s" % srcKey)

  dstKey = 'test-customer-encryption.cp'
  copy_customer_encryption(dstKey, srcKey, headers={
    **customer_encryption_headers,
    'x-kss-copy-source-server-side-encryption-customer-algorithm': 'AES256',
    'x-kss-copy-source-server-side-encryption-customer-key': customer_key_base64,
    'x-kss-copy-source-server-side-encryption-customer-key-MD5': customer_key_md5
  })
  print("复制成功: %s" % dstKey)
  getObjectMeta(bucket_name, dstKey, headers=customer_encryption_headers)
  downloadObjectAndPrint(bucket_name, dstKey, headers=customer_encryption_headers)

  bucket.delete_key(srcKey)
  print("清理文件成功: %s" % srcKey)
  bucket.delete_key(dstKey)
  print("清理文件成功: %s" % dstKey)

# 服务端加密最终的文件
def test_multipart_upload_sse():
  import math, os
  from filechunkio import FileChunkIO
  bucket = conn.get_bucket(bucket_name)

  source_path = '/Users/jabbar/Downloads/article_cp.txt'
  # 源文件大小
  source_size = os.stat(source_path).st_size

  mp = bucket.initiate_multipart_upload("article_cp.txt.mpu.1", headers={
    # 'x-kss-server-side-encryption-customer-algorithm': 'AES256',
    # 'x-kss-server-side-encryption-customer-key': customer_key_base64,
    # 'x-kss-server-side-encryption-customer-key-MD5': customer_key_md5
  })
  print(mp)

  # chunk_size = 5242880
  chunk_size = 5242880
  chunk_count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

  # # 通过 FileChunkIO 将文件分片
  # for i in range(chunk_count):
  #   offset = chunk_size * i
  #   bytes = min(chunk_size, source_size - offset)
  #   with FileChunkIO(source_path, 'r', offset=offset, bytes=bytes) as fp:
  #     # 逐个上传分片
  #     mp.upload_part_from_file(fp, part_num=i + 1, headers={
  #       # 'x-kss-server-side-encryption-customer-algorithm': 'AES256',
  #       # 'x-kss-server-side-encryption-customer-key': customer_key_base64,
  #       # 'x-kss-server-side-encryption-customer-key-MD5': customer_key_md5
  #     })
  with open(source_path, 'rb') as fp:
    mp.upload_part_from_file(fp, part_num=1)
  # 发送请求，合并分片，完成分片上传
  ret = mp.complete_upload()
  if ret and ret.status == 200:
    print("上传成功")

# 以存在的 服务端加密的 object作为part，进行分块的合并
def test_multipart_upload_copy_with_sse_s3_object():
  bucket = conn.get_bucket(bucket_name)

  # uploadObjectFromFile('/Users/jabbar/tools/test-data/xaa', encrypt_key=True)
  # uploadObjectFromFile('/Users/jabbar/tools/test-data/xab', encrypt_key=True)
  # uploadObjectFromFile('/Users/jabbar/tools/test-data/xac', encrypt_key=True)

  final_file = "article_cp.txt"
  # 测试加密最终对象
  mp = bucket.initiate_multipart_upload(final_file, encrypt_key=True, headers={
    # 'x-kss-server-side-encryption-customer-algorithm': 'AES256',
    # 'x-kss-server-side-encryption-customer-key': customer_key_base64,
    # 'x-kss-server-side-encryption-customer-key-MD5': customer_key_md5
  })
  print(mp.id)

  # mp.copy_part_from_key(bucket_name, 'DSC03380.ssec.jpg', 1, headers={
  #   # 'x-kss-server-side-encryption-customer-algorithm': 'AES256',
  #   # 'x-kss-server-side-encryption-customer-key': customer_key_base64,
  #   # 'x-kss-server-side-encryption-customer-key-MD5': customer_key_md5,
  #   'x-kss-copy-source-server-side-encryption-customer-algorithm': 'AES256',
  #   'x-kss-copy-source-server-side-encryption-customer-key': customer_key_base64,
  #   'x-kss-copy-source-server-side-encryption-customer-key-MD5': customer_key_md5
  # })
  # parts = mp.get_all_parts()
  # print(parts[0].size)
  # mp.copy_part_from_key(bucket_name, 'xaa', 1)
  # mp.copy_part_from_key(bucket_name, 'xab', 2)
  # mp.copy_part_from_key(bucket_name, 'xac', 3)

  # 发送请求，合并分片，完成分片上传
  ret = mp.complete_upload()
  if ret and ret.status == 200:
    print("上传成功: %s"% final_file)

  # getObjectMeta(bucket_name, final_file)
  # file_save_path = '/Users/jabbar/tools/test-data/' + final_file
  # downloadObjectAndSave(final_file, file_save_path)
  # print("下载文件成功")


  # bucket.delete_key(final_file)
  # print("清理文件成功: %s" % final_file)

# 以存在的object作为part，进行分块的合并
# 用户指定秘钥的服务端加密
def test_multipart_upload_copy_with_ss3_c_object():
  bucket = conn.get_bucket(bucket_name)

  customer_key = bytes(32)
  import base64, hashlib
  customer_key_base64 = base64.b64encode(customer_key).decode('utf-8')
  md5 = hashlib.md5()
  md5.update(customer_key)
  customer_key_md5 = base64.b64encode(md5.digest()).decode('utf-8')
  uploadObjectFromFile('/Users/jabbar/tools/test-data/xaa', headers=customer_encryption_headers)
  uploadObjectFromFile('/Users/jabbar/tools/test-data/xab', headers=customer_encryption_headers)
  uploadObjectFromFile('/Users/jabbar/tools/test-data/xac', headers=customer_encryption_headers)

  final_file = "DSC03380.encrypt.customer.jpg"
  # 测试加密最终对象
  mp = bucket.initiate_multipart_upload(final_file, headers=customer_encryption_headers)
  print(mp)

  mp.copy_part_from_key(bucket_name, 'xaa', 1, headers={
    **customer_encryption_headers,
    'x-kss-copy-source-server-side-encryption-customer-algorithm': 'AES256',
    'x-kss-copy-source-server-side-encryption-customer-key': customer_key_base64,
    'x-kss-copy-source-server-side-encryption-customer-key-MD5': customer_key_md5
  })
  mp.copy_part_from_key(bucket_name, 'xab', 2, headers={
    **customer_encryption_headers,
    'x-kss-copy-source-server-side-encryption-customer-algorithm': 'AES256',
    'x-kss-copy-source-server-side-encryption-customer-key': customer_key_base64,
    'x-kss-copy-source-server-side-encryption-customer-key-MD5': customer_key_md5
  })
  mp.copy_part_from_key(bucket_name, 'xac', 3, headers={
    **customer_encryption_headers,
    'x-kss-copy-source-server-side-encryption-customer-algorithm': 'AES256',
    'x-kss-copy-source-server-side-encryption-customer-key': customer_key_base64,
    'x-kss-copy-source-server-side-encryption-customer-key-MD5': customer_key_md5
  })

  # 发送请求，合并分片，完成分片上传
  ret = mp.complete_upload()
  if ret and ret.status == 200:
    print("上传成功: %s"% final_file)

  getObjectMeta(bucket_name, final_file, headers=customer_encryption_headers)
  file_save_path = '/Users/jabbar/tools/test-data/' + final_file
  downloadObjectAndSave(final_file, file_save_path, headers=customer_encryption_headers)
  print("下载文件成功")
  bucket.delete_key(final_file)
  print("清理文件成功: %s" % final_file)




# test_multipart_upload_copy_with_encrypt_object()
# getObjectMeta(bucket_name, 'DSC03380.encrypt.jpg')
# copy_encryption('ks3util_bak', 'ks3util', encrypt_key=True)
# uploadObjectFromFile('/Users/jabbar/Downloads/DSC03380.encrypt.jpg', file_name='DSC03380.ssec.jpg', headers=customer_encryption_headers)
# getObjectMeta('test-bucket', 'DSC03380.ssec.jpg', headers=customer_encryption_headers)
# uploadObjectFromFile('/Users/jabbar/Downloads/article_cp.txt', file_name='article_cp.txt_ss3_s3', encrypt_key=True)
# getObjectMeta(bucket_name, 'article_cp.txt_ss3_s3', headers={
#   'x-kss-server-side-encryption': 'AES256'
# })
# test_copy_customer_encryption()
# upload_object_with_customer_encryption_key('test-customer-encryption', 'helloworld')
# test_multipart_upload_copy_with_customer_encrypt_object()


import logging
import time

from ks3.utils import RetryPolicy, ExponentialWait, StopAfterAttempt

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2

logger = logging.getLogger(__name__)


class _BaseListResultSet(object):
    """
    A base class for listing items which can be configured with retry and request interval settings.
    """
    def __init__(self, retry_policy=RetryPolicy(), request_interval=0):
        self.retry_policy = retry_policy
        if retry_policy is None:
            self.retry_policy = RetryPolicy(None, None)
        self.request_interval = request_interval

    def auto_list(self, fn, **params):
        """
        Automatic list using specific methods.

        :param fn: specific method for listing items.
        :param params: parameters for fn.
        """
        more_results = True
        k = None
        cursor_dict = self.init_cursor()
        while more_results:
            params.update(cursor_dict)

            rs = fn(**params)
            # rs = self.retry_policy.call(fn, **params)

            start_time = time.time()
            for k in rs:
                yield k
            cursor_dict = self.update_cursor(rs, k)
            more_results = rs.is_truncated
            if more_results:
                elapsed_time = time.time() - start_time
                if elapsed_time < self.request_interval:
                    time.sleep(self.request_interval - elapsed_time)

    def init_cursor(self):
        """
        Implement this method to initially obtain paging cursor parameters for listing.
        'Paging cursor parameters' such as marker, continuation_token, etc.

        :return: dict[str, str]
        """
        raise NotImplementedError

    def update_cursor(self, rs, k):
        """
        Implement this method to update paging cursor parameters for listing.
        so that the next request can pick up where the previous one left off.

        :param rs: the result set returned by the request.
        :param k: the last key returned by the request.
        :return: dict[str, str]
        """
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError


class BucketListResultSet(_BaseListResultSet):
    """
    A result set for listing keys within a bucket.  This transparently handles the results paging from S3
    so even if you have many thousands of keys within the bucket you can iterate over all
    keys in a reasonably efficient manner.
    """

    def __init__(self, bucket=None, prefix='', delimiter='', marker='', max_keys='', encoding_type='',
                 continuation_token='', fetch_owner='', list_type=None, start_after='',
                 retry_policy=RetryPolicy(), request_interval=0):
        super().__init__(retry_policy, request_interval)
        self.bucket = bucket
        self.prefix = prefix
        self.delimiter = delimiter
        self.marker = marker
        self.max_keys = max_keys
        self.encoding_type = encoding_type
        self.continuation_token = continuation_token
        self.fetch_owner = fetch_owner
        self.list_type = list_type
        self.start_after = start_after

    def init_cursor(self):
        return {'marker': self.marker, 'continuation_token': self.continuation_token}

    def update_cursor(self, rs, k):
        marker = ''
        continuation_token = ''
        if rs.next_continuation_token:
            continuation_token = rs.next_continuation_token
        elif k:
            marker = rs.next_marker or k.name
        if marker:
            marker = parse.unquote(marker)
        return {'marker': marker, 'continuation_token': continuation_token}

    def __iter__(self):
        return self.auto_list(self.bucket.get_all_keys, retry_policy=self.retry_policy, prefix=self.prefix, delimiter=self.delimiter,
                              max_keys=self.max_keys, encoding_type=self.encoding_type,
                              fetch_owner=self.fetch_owner, list_type=self.list_type, start_after=self.start_after)


class VersionedBucketListResultSet(_BaseListResultSet):
    """
    A result set for listing versions within a bucket. This transparently handles the results paging from S3
    so even if you have many thousands of keys within the bucket you can iterate over all
    keys in a reasonably efficient manner.
    """

    def __init__(self, bucket=None, prefix='', delimiter='', key_marker='', version_id_marker='', headers=None,
                 encoding_type=None, retry_policy=RetryPolicy(),
                 request_interval=0):
        super().__init__(retry_policy, request_interval)
        self.bucket = bucket
        self.prefix = prefix
        self.delimiter = delimiter
        self.key_marker = key_marker
        self.version_id_marker = version_id_marker
        self.headers = headers
        self.encoding_type = encoding_type

    def init_cursor(self):
        return {'key_marker': self.key_marker}

    def update_cursor(self, rs, k):
        key_marker = rs.next_key_marker
        if key_marker:
            key_marker = parse.unquote(key_marker)
        return {'key_marker': key_marker}

    def __iter__(self):
        return self.auto_list(self.bucket.get_all_versions, prefix=self.prefix, delimiter=self.delimiter,
                              key_marker=self.key_marker, version_id_marker=self.version_id_marker,
                              headers=self.headers, encoding_type=self.encoding_type)


class BucketRetentionListResultSet(_BaseListResultSet):
    def __init__(self, bucket=None, prefix='', delimiter='', marker='', max_keys='',
                 retry_policy=RetryPolicy(), request_interval=0):
        super().__init__(retry_policy, request_interval)
        self.bucket = bucket
        self.prefix = prefix
        self.delimiter = delimiter
        self.marker = marker
        self.max_keys = max_keys

    def init_cursor(self):
        return {'marker': self.marker}

    def update_cursor(self, rs, k):
        marker = ''
        if k:
            marker = rs.next_marker or k.name
        if marker:
            marker = parse.unquote(marker)
        return {'marker': marker}

    def __iter__(self):
        return self.auto_list(self.bucket.get_all_retention_keys, prefix=self.prefix, delimiter=self.delimiter,
                              marker=self.marker, max_keys=self.max_keys)


class MultiPartUploadListResultSet(_BaseListResultSet):
    """
    A result set for listing multipart uploads within a bucket.
    This transparently handles the results paging from S3 so even if you have
    many thousands of uploads within the bucket you can iterate over all
    keys in a reasonably efficient manner.
    """

    def __init__(self, bucket=None, key_marker='', upload_id_marker='', headers=None, encoding_type=None,
                 retry_policy=RetryPolicy(), request_interval=0):
        super().__init__(retry_policy, request_interval)
        self.bucket = bucket
        self.key_marker = key_marker
        self.upload_id_marker = upload_id_marker
        self.headers = headers
        self.encoding_type = encoding_type

    def init_cursor(self):
        return {'key_marker': self.key_marker, 'upload_id_marker': self.upload_id_marker}

    def update_cursor(self, rs, k):
        return {'key_marker': rs.next_key_marker, 'upload_id_marker': rs.next_upload_id_marker}

    def __iter__(self):
        return self.auto_list(self.bucket.get_all_multipart_uploads, key_marker=self.key_marker,
                              upload_id_marker=self.upload_id_marker, headers=self.headers,
                              encoding_type=self.encoding_type)



#!/usr/bin/env python

"""
distutils/setuptools install script.
"""

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import re

sdk_version = ''
with open('ks3/__init__.py', 'r') as fd:
    sdk_version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)


with open('README.md', 'rb') as f:
    readme = f.read().decode('utf-8')

setup(
    name='ks3sdk',
    version=sdk_version,
    description='Ksyun KS3 SDK',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=['ks3', 'ks3.xmlParsers'],
    install_requires=[
        'six', 'python-dateutil',
        'aiofiles~=23.2.1',
        'crcmod~=1.7',
        'pycryptodome~=3.20.0',
        'requests',
    ],
    include_package_data=True,
    url='https://gitee.com/ks3sdk/ks3-python-sdk',
    author='ksc_ks3',
    author_email='ksc_ks3@kingsoft.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    python_requires='>=3.6'
)



import logging
import os
import math
from concurrent import futures

from ks3.exception import KS3ClientError
from ks3.multipart import MultiPartUpload, PartInfo
from ks3.utils import ChunkIO, ResumeRecordManager

KB = 1024
MB = KB * KB
GB = KB * MB

MAX_PARTS = 10_000
MAX_SIZE = 5 * GB
MIN_SIZE = 100 * KB

logger = logging.getLogger(__name__)


class UploadTask:
    def __init__(
            self, key, bucket, filename,
            executor,
            part_size=100 * KB,
            resumable=False,
            resumable_filename=None,
    ):
        """
        key: 需要上传的Key对象（ks3.key.Key）
        bucket: 需要上传到的Bucket对象（ks3.bucket.Bucket）
        filename: 待上传的本地文件路径
        executor: 线程池
        part_size: 分块大小，默认100KB。小于分块大小的文件，使用简单上传；大于分块大小的文件，使用分块上传
        resumable: 是否开启断点续传，默认False。开启断点续传时，如果本地存在有效的断点续传记录，则尝试恢复上传，否则初始化分块上传
        resumable_filename: 断点续传记录文件路径。如果不指定，则使用默认路径（self.filename + '.ks3resume'）
        """
        self.key = key
        self.bucket = bucket
        self.filename = filename
        self.executor = executor
        self.part_size = part_size
        self.resumable = resumable
        if self.resumable:
            if resumable_filename is None:
                resumable_filename = self.filename + '.ks3resume'
            self.record_manager = ResumeRecordManager(resumable_filename)
        else:
            self.record_manager = None
        # 本地文件的修改时间
        self.__mtime = os.path.getmtime(filename)
        # 本地文件的大小
        self.__file_size = os.path.getsize(filename)

    def upload(self, headers=None):
        if self.__file_size <= self.part_size:
            logger.debug('key_name={0}, file_size({1}) less than part_size({2}), use simple upload'
                         .format(self.key.name, self.__file_size, self.part_size))
            self.key.set_contents_from_filename(self.filename, headers=headers)
        else:
            logger.debug('key_name={0}, file_size({1}) greater than part_size({2}), use multipart upload'
                         .format(self.key.name, self.__file_size, self.part_size))
            self.multipart_upload(headers)

    def multipart_upload(self, headers=None):
        # 初始化分块上传
        mp = self._get_or_init_record_as_mp(headers)

        if not self.resumable:
            part_uploaded = set()
        elif self.bucket.connection.enable_crc:
            # 开启了crc校验，已上传分块以本地记录为准，防止本地与远端记录分块数据不一致。
            part_uploaded = set(self.record_manager.record.part_crc_infos.keys())
        else:
            # 未开启crc校验，已上传分块以远端记录为准（通过断点续传文件内记录的upload_id等信息获取远端分块记录）。
            part_uploaded = set()
            # for循环将调用 List Parts 接口
            for p in mp:
                part_uploaded.add(p.part_number)

        part_size = self._adjust_part_size()
        num_parts = int(math.ceil(self.__file_size / float(part_size)))
        remain_size = self.__file_size
        # 分块上传
        # part_futures<future, part_num>
        part_futures = {}
        for part_num in range(1, num_parts + 1):
            # 跳过已上传的part
            if part_num in part_uploaded:
                logger.debug('key_name={0}, part {1} already uploaded, skip'.format(self.key.name, part_num))
                remain_size -= part_size
                continue
            part_futures[self.executor.submit(
                self._upload_part,
                mp=mp,
                part_num=part_num,
                part_size=min(part_size, remain_size),
                headers=headers,
            )] = part_num
            remain_size -= part_size
        # 不能及时处理中断信号
        self.executor.shutdown(wait=True)
        for future in futures.as_completed(part_futures):
            try:
                future.result()
            except Exception as e:
                logger.error('key_name={0}, part {1} upload failed: {2}'.format(self.key.name, part_futures[future], e))
                raise e
        try:
            resp = mp.complete_upload()
        except KS3ClientError as e:
            # crc校验失败，删除本地断点续传记录，以便于下次重新上传
            if 'Inconsistent CRC checksum' in e.reason:
                if self.resumable:
                    self.record_manager.delete()
            raise e
        # 正常上传完成，未抛出异常，则删除本地断点续传记录
        if self.resumable:
            self.record_manager.delete()
        return resp

    def _get_or_init_record_as_mp(self, headers):
        """
        获取或初始化分块上传:
        1. 如果没有开启断点续传，则初始化分块上传
        2. 如果开启断点续传，则尝试加载本地记录，如果记录存在且有效，则恢复分块上传，否则初始化分块上传
        return: 返回分块上传对象MultiPartUpload类型
        """
        if not self.resumable:
            mp = self.bucket.initiate_multipart_upload(self.key.name, headers=headers)
            return mp

        self.record_manager.load()

        if self.record_manager.record is not None and not self._check_record_valid():
            logger.debug('key_name={0}, found invalid record, delete it'.format(self.key.name))
            self.record_manager.delete()
        if self.record_manager.record is None:
            logger.debug('key_name={0}, not found record, initiate multipart upload'.format(self.key.name))
            mp = self.bucket.initiate_multipart_upload(self.key.name, headers=headers)
            self.record_manager.record = UploadRecord(mp.id, self.__file_size, self.__mtime,
                                                      self.bucket.name, self.key.name, self.part_size, {})
            self.record_manager.save()
        else:
            logger.debug('key_name={0}, found record, resume multipart upload: {1}'
                         .format(self.key.name, self.record_manager.record.upload_id))
            mp = MultiPartUpload(self.bucket)
            mp.id = self.record_manager.record.upload_id
            mp.key_name = self.record_manager.record.key_name
            mp.part_crc_infos = self.record_manager.record.part_crc_infos
            self.part_size = self.record_manager.record.part_size
        return mp

    def _check_record_valid(self):
        record = self.record_manager.record
        if not isinstance(record, UploadRecord):
            return False
        for attr in [record.upload_id, record.bucket_name, record.key_name]:
            if not isinstance(attr, str):
                return False
        for attr in [record.file_size, record.part_size]:
            if not isinstance(attr, int):
                return False
        if not isinstance(record.mtime, float) and not isinstance(record.mtime, int):
            return False
        if not isinstance(record.part_crc_infos, dict):
            return False
        if record.mtime != self.__mtime or record.file_size != self.__file_size:
            return False
        return True

    def _adjust_part_size(self):
        part_size = self.part_size
        num_parts = int(math.ceil(self.__file_size / float(part_size)))
        # 以分块数的标准，调整分块大小
        while num_parts > MAX_PARTS:
            part_size *= 2
            num_parts = int(math.ceil(self.__file_size / float(part_size)))
        # 以单次上传大小的标准，调整分块大小
        if part_size > MAX_SIZE:
            part_size = MAX_SIZE
        elif part_size < MIN_SIZE:
            part_size = MIN_SIZE
        return part_size

    def _upload_part(self, mp, part_num, part_size, headers=None):
        with ChunkIO(self.filename, (part_num - 1) * self.part_size, part_size) as fp:
            resp = mp.upload_part_from_file(fp, part_num, headers=headers)
            if self.resumable and self.bucket.connection.enable_crc:
                self.record_manager.record.part_crc_infos[part_num] = PartInfo(
                    part_size, resp.getheader(self.bucket.connection.provider.checksum_crc64ecma_header))
                # 保存crc信息
                self.record_manager.save()


class UploadRecord(object):
    def __init__(self, upload_id, file_size, mtime, bucket_name, key_name, part_size, part_crc_infos):
        self.upload_id = upload_id
        self.file_size = file_size
        self.mtime = mtime
        self.bucket_name = bucket_name
        self.key_name = key_name
        self.part_size = part_size
        self.part_crc_infos = part_crc_infos
